try:
    from StringIO import StringIO  # Python 2
except ImportError:
    from io import StringIO  # Python 3

import os
import shutil
import tempfile

try:
    from unittest import mock
except ImportError:
    import mock

from django.core.management import call_command
from django.template import Context, Engine, Template
from django.test import TestCase

from flexi_tag.management.commands.generate_tag_models import Command
from flexi_tag.utils.models import FlexiTagMixin


class TagModelGenerationTestCase(TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

        self.model_template_string = """
# This file is auto-generated. Do not edit manually.
# Generated by the generate_tag_models command.

from django.db import models
from flexi_tag.utils.compat import JSONField, GinIndex


class {{ model_name }}Tag(models.Model):
    instance = models.OneToOneField(
        "{{ app_label }}.{{ model_name }}",
        on_delete=models.CASCADE,
        primary_key=True,
    )
    tags = JSONField(default=list)

    class Meta:
        indexes = [GinIndex(fields=["tags"])]

    def __str__(self):
        return "Tags for {}".format(self.instance)"
"""

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def _create_mock_model(
        self, name="MyModel", app_label="myapp", module="myapp.models"
    ):
        mock_model = mock.MagicMock()
        mock_model.__name__ = name
        mock_model._meta.app_label = app_label
        mock_model.__module__ = module
        mock_model.__bases__ = (FlexiTagMixin,)
        return mock_model


class GenerateTagModelsCommandTestCase(TagModelGenerationTestCase):
    @mock.patch("django.template.Engine.get_default")
    def test_template_content(self, mock_get_default):
        engine = Engine()
        mock_get_default.return_value = engine

        template = Template(self.model_template_string)

        context = Context(
            {
                "model_name": "TaggableManagerTestModel",
                "app_label": "testapp",
                "model_lower_name": "taggablemanagertestmodel",
            }
        )

        rendered_template = template.render(context)

        self.assertIn(
            "class TaggableManagerTestModelTag(models.Model):", rendered_template
        )

    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    def test_command_with_no_flexi_tag_models(
        self, mock_template, mock_get_app_configs
    ):
        mock_template_instance = mock.MagicMock()
        mock_template_instance.render.return_value = "# Generated model content"
        mock_template.return_value = mock_template_instance

        out = StringIO()

        mock_model = mock.MagicMock()
        mock_model.__name__ = "RegularModel"
        mock_model.__bases__ = (object,)  # Not a FlexiTagMixin

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model]
        mock_get_app_configs.return_value = [mock_app_config]

        call_command("generate_tag_models", stdout=out)

        self.assertIn("Successfully generated all tag models", out.getvalue())

    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    @mock.patch("flexi_tag.management.commands.generate_tag_models.import_module")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.inspect.getfile")
    def test_dry_run_option(
        self,
        mock_getfile,
        mock_import_module,
        mock_get_app_configs,
        mock_template_class,
    ):
        mock_template = mock.MagicMock()
        mock_template.render.return_value = "# Mocked template content"
        mock_template_class.return_value = mock_template

        mock_model = self._create_mock_model(
            name="TaggableManagerTestModel",
            app_label="testapp",
            module="testapp.models",
        )

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model]
        mock_get_app_configs.return_value = [mock_app_config]

        mock_module = mock.MagicMock()
        mock_import_module.return_value = mock_module
        test_file_path = os.path.join(self.test_dir, "models.py")
        mock_getfile.return_value = test_file_path

        # Run the command with --dry-run
        out = StringIO()
        call_command("generate_tag_models", "--dry-run", stdout=out)

        output = out.getvalue()
        self.assertIn("Would create file:", output)
        self.assertIn("Dry run completed. No files were created.", output)
        self.assertNotIn("Successfully generated all tag models", output)

    @mock.patch("sys.exit")
    def test_command_help(self, mock_exit):
        command = Command()
        command.stdout = StringIO()

        help_text = command.help

        self.assertIn("Generate tag models", help_text)
        self.assertTrue(callable(command.add_arguments))

        parser = mock.MagicMock()
        command.add_arguments(parser)
        self.assertTrue(parser.add_argument.called)
        args, _ = parser.add_argument.call_args
        self.assertEqual(args[0], "--dry-run")

        mock_exit.assert_not_called()


class CustomTagGenerationTestCase(TagModelGenerationTestCase):
    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    @mock.patch("flexi_tag.management.commands.generate_tag_models.import_module")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.inspect.getfile")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.open", create=True)
    def test_multiple_models_generation(
        self,
        mock_open,
        mock_getfile,
        mock_import_module,
        mock_get_app_configs,
        mock_template_class,
    ):
        mock_template = mock.MagicMock()
        mock_template.render.return_value = "# Mocked template content"
        mock_template_class.return_value = mock_template

        mock_model1 = self._create_mock_model(name="Model1", app_label="testapp")
        mock_model2 = self._create_mock_model(name="Model2", app_label="testapp")

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model1, mock_model2]
        mock_get_app_configs.return_value = [mock_app_config]

        mock_module = mock.MagicMock()
        mock_import_module.return_value = mock_module
        test_file_path = os.path.join(self.test_dir, "models.py")
        mock_getfile.return_value = test_file_path

        mock_file = mock.MagicMock()
        mock_open.return_value.__enter__.return_value = mock_file

        out = StringIO()
        call_command("generate_tag_models", stdout=out)

        self.assertEqual(mock_open.call_count, 2)
        self.assertEqual(mock_file.write.call_count, 2)
        self.assertIn("Successfully generated all tag models", out.getvalue())

    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    @mock.patch("flexi_tag.management.commands.generate_tag_models.import_module")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.inspect.getfile")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.open", create=True)
    def test_model_with_special_name(
        self,
        mock_open,
        mock_getfile,
        mock_import_module,
        mock_get_app_configs,
        mock_template_class,
    ):
        mock_template = mock.MagicMock()
        mock_template.render.return_value = "# Mocked template content"
        mock_template_class.return_value = mock_template

        special_name = "Special_Model_Name-With.Characters$"
        mock_model = self._create_mock_model(name=special_name, app_label="testapp")

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model]
        mock_get_app_configs.return_value = [mock_app_config]

        mock_module = mock.MagicMock()
        mock_import_module.return_value = mock_module
        test_file_path = os.path.join(self.test_dir, "models.py")
        mock_getfile.return_value = test_file_path

        mock_file = mock.MagicMock()
        mock_open.return_value.__enter__.return_value = mock_file

        out = StringIO()
        call_command("generate_tag_models", stdout=out)

        self.assertEqual(mock_open.call_count, 1)
        self.assertEqual(mock_file.write.call_count, 1)
        self.assertEqual(mock_template.render.call_count, 1)

        context = mock_template.render.call_args[0][0]
        self.assertEqual(context["model_name"], special_name)
        self.assertEqual(context["model_lower_name"], special_name.lower())
        self.assertIn("Successfully generated all tag models", out.getvalue())

    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    @mock.patch("flexi_tag.management.commands.generate_tag_models.import_module")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.inspect.getfile")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.open", create=True)
    def test_command_with_non_django_class(
        self,
        mock_open,
        mock_getfile,
        mock_import_module,
        mock_get_app_configs,
        mock_template_class,
    ):
        mock_template = mock.MagicMock()
        mock_template.render.return_value = "# Mocked template content"
        mock_template_class.return_value = mock_template

        mock_model = mock.MagicMock()
        mock_model.__name__ = "NonModelClass"
        mock_model.__bases__ = (FlexiTagMixin,)
        mock_model._meta = mock.MagicMock()
        mock_model._meta.app_label = "testapp"
        mock_model.__module__ = "testapp.models"

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model]
        mock_get_app_configs.return_value = [mock_app_config]

        mock_module = mock.MagicMock()
        mock_import_module.return_value = mock_module
        test_file_path = os.path.join(self.test_dir, "models.py")
        mock_getfile.return_value = test_file_path

        mock_file = mock.MagicMock()
        mock_open.return_value.__enter__.return_value = mock_file

        out = StringIO()
        call_command("generate_tag_models", stdout=out)

        self.assertIn("Successfully generated all tag models", out.getvalue())


class AutoImportTestCase(TagModelGenerationTestCase):
    def setUp(self):
        super(AutoImportTestCase, self).setUp()
        self.models_content = """
from django.db import models
from flexi_tag.utils.models import FlexiTagMixin

class TaggableManagerTestModel(FlexiTagMixin):
    name = models.CharField(max_length=100)
"""

    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.inspect.getfile")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.import_module")
    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    def test_auto_import_single_model(
        self,
        mock_get_app_configs,
        mock_import_module,
        mock_getfile,
        mock_template_class,
    ):
        model_dir = os.path.join(self.test_dir, "testapp")
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)
        models_path = os.path.join(model_dir, "models.py")

        with open(models_path, "w") as f:
            f.write(self.models_content)

        mock_template = mock.MagicMock()
        mock_template.render.return_value = "# Generated model content"
        mock_template_class.return_value = mock_template

        mock_model = self._create_mock_model(
            name="TaggableManagerTestModel", app_label="testapp"
        )

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model]
        mock_get_app_configs.return_value = [mock_app_config]

        mock_import_module.return_value = mock.MagicMock()
        mock_getfile.return_value = os.path.join(model_dir, "__init__.py")

        out = StringIO()
        call_command("generate_tag_models", stdout=out)

        with open(models_path, "r") as f:
            updated_content = f.read()

        self.assertIn(
            "from .flexi_generated_model import TaggableManagerTestModelTag  # noqa",
            updated_content,
        )

    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.inspect.getfile")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.import_module")
    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    def test_auto_import_multiple_models(
        self,
        mock_get_app_configs,
        mock_import_module,
        mock_getfile,
        mock_template_class,
    ):
        models_content = """
from django.db import models
from flexi_tag.utils.models import FlexiTagMixin

class FirstModel(FlexiTagMixin):
    name = models.CharField(max_length=100)

class SecondModel(FlexiTagMixin):
    title = models.CharField(max_length=200)
"""

        model_dir = os.path.join(self.test_dir, "testapp")
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)
        models_path = os.path.join(model_dir, "models.py")

        with open(models_path, "w") as f:
            f.write(models_content)

        mock_template = mock.MagicMock()
        mock_template.render.return_value = "# Generated model content"
        mock_template_class.return_value = mock_template

        mock_model1 = self._create_mock_model(name="FirstModel", app_label="testapp")
        mock_model2 = self._create_mock_model(name="SecondModel", app_label="testapp")

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model1, mock_model2]
        mock_get_app_configs.return_value = [mock_app_config]

        mock_import_module.return_value = mock.MagicMock()
        mock_getfile.return_value = os.path.join(model_dir, "__init__.py")

        out = StringIO()
        call_command("generate_tag_models", stdout=out)

        with open(models_path, "r") as f:
            updated_content = f.read()

        self.assertIn("from .flexi_generated_model import", updated_content)
        self.assertIn("FirstModelTag", updated_content)
        self.assertIn("SecondModelTag", updated_content)
        self.assertIn("# noqa", updated_content)
